# -*- coding: utf-8 -*-
from pandas_ta import alma
from pandas_ta import dema
from pandas_ta import ema
from pandas_ta import fwma
from pandas_ta import hilo
from pandas_ta import hl2
from pandas_ta import hlc3
from pandas_ta import hma
from pandas_ta import hwma
from pandas_ta import ichimoku
from pandas_ta import jma
from pandas_ta import kama
from pandas_ta import linreg
from pandas_ta import ma
from pandas_ta import mcgd
from pandas_ta import midpoint
from pandas_ta import midprice
from pandas_ta import ohlc4
from pandas_ta import pwma
from pandas_ta import rma
from pandas_ta import sinwma
from pandas_ta import sma
from pandas_ta import ssf
from pandas_ta import supertrend
from pandas_ta import swma
from pandas_ta import t3
from pandas_ta import tema
from pandas_ta import trima
from pandas_ta import vidya
from pandas_ta import vwap
from pandas_ta import vwma
from pandas_ta import wcp
from pandas_ta import wma
from pandas_ta import zlma
